# Copyright (c) 2019-2020 Phase Advanced Sensor Systems, Inc.
from psdb.devices import Reg32, Reg32R, Reg32W
from . import scs_base

import collections


# The core registers and the selector value they map to in the DCRSR.
CORE_REGISTERS = collections.OrderedDict([
    ('r0',      0),
    ('r1',      1),
    ('r2',      2),
    ('r3',      3),
    ('r4',      4),
    ('r5',      5),
    ('r6',      6),
    ('r7',      7),
    ('r8',      8),
    ('r9',      9),
    ('r10',     10),
    ('r11',     11),
    ('r12',     12),
    ('sp',      13),
    ('lr',      14),
    ('pc',      15),
    ('xpsr',    16),
    ('msp',     17),
    ('psp',     18),
    ('cfbp',    20),   # CONTROL, FAULTMASK, BASEPRI, PRIMASK
    ('fpscr',   33),
    ('s0',      64),
    ('s1',      65),
    ('s2',      66),
    ('s3',      67),
    ('s4',      68),
    ('s5',      69),
    ('s6',      70),
    ('s7',      71),
    ('s8',      72),
    ('s9',      73),
    ('s10',     74),
    ('s11',     75),
    ('s12',     76),
    ('s13',     77),
    ('s14',     78),
    ('s15',     79),
    ('s16',     80),
    ('s17',     81),
    ('s18',     82),
    ('s19',     83),
    ('s20',     84),
    ('s21',     85),
    ('s22',     86),
    ('s23',     87),
    ('s24',     88),
    ('s25',     89),
    ('s26',     90),
    ('s27',     91),
    ('s28',     92),
    ('s29',     93),
    ('s30',     94),
    ('s31',     95),
])


class SCS(scs_base.SCS):
    '''
    Driver for Cortex V7-M (M4, M7) System Control Space.
    '''
    REGS = [Reg32 ('MCR',    0x000),
            Reg32R('ICTR',   0x004, [('INTLINESNUM', 4)]),
            Reg32 ('ACTLR',  0x008),
            Reg32 ('SYST_CSR',   0x010, [('ENABLE',     1),
                                         ('TICKINT',    1),
                                         ('CLKSOURCE',  1),
                                         ('',          13),
                                         ('COUNTFLAG',  1),
                                         ]),
            Reg32 ('SYST_RVR',   0x014, [('RELOAD', 24)]),
            Reg32 ('SYST_CVR',   0x018, [('CURRENT', 24)]),
            Reg32 ('SYST_CALIB', 0x01C, [('TENMS',     24),
                                         ('',           6),
                                         ('SKEW',       1),
                                         ('NOREF',      1),
                                         ]),
            Reg32R('CPUID',  0xD00, [('REVISION',    4),
                                     ('PARTNO',     12),
                                     ('ONES',        4),
                                     ('VARIANT',     4),
                                     ('IMPLEMENTER', 8),
                                     ]),
            Reg32 ('ICSR',   0xD04, [('VECTACTIVE',  9),
                                     ('',            2),
                                     ('RETTOBASE',   1),
                                     ('VECTPENDING', 9),
                                     ('',            1),
                                     ('ISRPENDING',  1),
                                     ('ISRPREEMPT',  1),
                                     ('',            1),
                                     ('PENDSTCLR',   1),
                                     ('PENDSTSET',   1),
                                     ('PENDSVCLR',   1),
                                     ('PENDSVSET',   1),
                                     ('',            2),
                                     ('NMIPENDSET',  1),
                                     ]),
            Reg32 ('VTOR',   0xD08, [('',        7),
                                     ('TBLOFF', 25),
                                     ]),
            Reg32 ('AIRCR',  0xD0C, [('VECTRESET',     1),
                                     ('VECTCLRACTIVE', 1),
                                     ('SYSRESETREQ',   1),
                                     ('',              5),
                                     ('PRIGROUP',      3),
                                     ('',              4),
                                     ('ENDIANNESS',    1),
                                     ('VECTKEYSTAT',  16),
                                     ]),
            Reg32 ('SCR',    0xD10, [('',              1),
                                     ('SLEEPONEXIT',   1),
                                     ('SLEEPDEEP',     1),
                                     ('',              1),
                                     ('SEVONPEND',     1),
                                     ]),
            Reg32 ('CCR',    0xD14, [('NONBASETHRDENA', 1),
                                     ('USERSETMPEND',   1),
                                     ('',               1),
                                     ('UNALIGN_TRP',    1),
                                     ('DIV_0_TRP',      1),
                                     ('',               3),
                                     ('BFHFNMIGN',      1),
                                     ('STKALIGN',       1),
                                     ('',               6),
                                     ('DC',             1),
                                     ('IC',             1),
                                     ('BP',             1),
                                     ]),
            Reg32 ('SHPR1',  0xD18, [('PRI_4', 8),
                                     ('PRI_5', 8),
                                     ('PRI_6', 8),
                                     ('PRI_7', 8),
                                     ]),
            Reg32 ('SHPR2',  0xD1C, [('PRI_8',  8),
                                     ('PRI_9',  8),
                                     ('PRI_10', 8),
                                     ('PRI_11', 8),
                                     ]),
            Reg32 ('SHPR3',  0xD20, [('PRI_12', 8),
                                     ('PRI_13', 8),
                                     ('PRI_14', 8),
                                     ('PRI_15', 8),
                                     ]),
            Reg32 ('SHCSR',  0xD24, [('MEMFAULTACT',    1),
                                     ('BUSFAULTACT',    1),
                                     ('',               1),
                                     ('USGFAULTACT',    1),
                                     ('',               3),
                                     ('SVCALLACT',      1),
                                     ('MONITORACT',     1),
                                     ('',               1),
                                     ('PENDSVACT',      1),
                                     ('SYSTICKACT',     1),
                                     ('USGFAULTPENDED', 1),
                                     ('MEMFAULTPENDED', 1),
                                     ('BUSFAULTPENDED', 1),
                                     ('SVCALLPENDED',   1),
                                     ('MEMFAULTENA',    1),
                                     ('BUSFAULTENA',    1),
                                     ('USGFAULTENA',    1),
                                     ]),
            Reg32 ('CFSR',   0xD28, [('IACCVIOL',    1),
                                     ('DACCVIOL',    1),
                                     ('',            1),
                                     ('MUNSTKERR',   1),
                                     ('MSTKERR',     1),
                                     ('MLSPERR',     1),
                                     ('',            1),
                                     ('MMARVALID',   1),
                                     ('IBUSERR',     1),
                                     ('PRECISERR',   1),
                                     ('IMPRECISERR', 1),
                                     ('UNSTKERR',    1),
                                     ('STKERR',      1),
                                     ('LSPERR',      1),
                                     ('',            1),
                                     ('BFARVALID',   1),
                                     ('UNDERINSTR',  1),
                                     ('INVSTATE',    1),
                                     ('INVPC',       1),
                                     ('NOCP',        1),
                                     ('',            4),
                                     ('UNALIGNED',   1),
                                     ('DIVBYZERO',   1),
                                     ('',            6),
                                     ]),
            Reg32 ('HFSR',   0xD2C, [('',            1),
                                     ('VECTTBL',     1),
                                     ('',           28),
                                     ('FORCED',      1),
                                     ('DEBUGEVT',    1),
                                     ]),
            Reg32 ('DFSR',   0xD30, [('HALTED',      1),
                                     ('BKPT',        1),
                                     ('DWTTRAP',     1),
                                     ('VCATCH',      1),
                                     ('EXTERNAL',    1),
                                     ]),
            Reg32 ('MMFAR',  0xD34, [('ADDRESS', 32)]),
            Reg32 ('BFAR',   0xD38, [('ADDRESS', 32)]),
            Reg32 ('AFSR',   0xD3C),
            Reg32 ('CPACR',  0xD88, [('CP0',  2),
                                     ('CP1',  2),
                                     ('CP2',  2),
                                     ('CP3',  2),
                                     ('CP4',  2),
                                     ('CP5',  2),
                                     ('CP6',  2),
                                     ('CP7',  2),
                                     ('',     4),
                                     ('CP10', 2),
                                     ('CP11', 2),
                                     ]),
            Reg32 ('DHCSR',  0xDF0, [('C_DEBUGEN',     1),
                                     ('C_HALT',        1),
                                     ('C_STEP',        1),
                                     ('C_MASKINTS',    1),
                                     ('',              1),
                                     ('C_SNAPSTALL',   1),
                                     ('',             10),
                                     ('S_REGRDY',      1),
                                     ('S_HALT',        1),
                                     ('S_SLEEP',       1),
                                     ('S_LOCKUP',      1),
                                     ('',              4),
                                     ('S_RETIRE_ST',   1),
                                     ('S_RESET_ST',    1),
                                     ]),
            Reg32W('DCRSR',  0xDF4, [('REGSEL',    7),
                                     ('',          9),
                                     ('REGWnR',    1),
                                     ]),
            Reg32 ('DCRDR',  0xDF8),
            Reg32 ('DEMCR',  0xDFC, [('VC_CORERESET', 1),
                                     ('',             3),
                                     ('VC_MMERR',     1),
                                     ('VC_NOCPERR',   1),
                                     ('VC_CHKERR',    1),
                                     ('VC_STATEERR',  1),
                                     ('VC_BUSERR',    1),
                                     ('VC_INTERR',    1),
                                     ('VC_HARDERR',   1),
                                     ('',             5),
                                     ('MON_EN',       1),
                                     ('MON_PEND',     1),
                                     ('MON_STEP',     1),
                                     ('MON_REQ',      1),
                                     ('',             4),
                                     ('TRCENA',       1),
                                     ]),
            ]

    def __init__(self, component, subtype):
        super(SCS, self).__init__(component, subtype, SCS.REGS, CORE_REGISTERS)

        # Enable DEMCR.TRCENA so we can probe further.
        self._DEMCR.TRCENA = 1
