# Copyright (c) 2018-2019 Phase Advanced Sensor Systems, Inc.
import time

from ..device import Device, AReg32


class PWR(Device):
    '''
    Driver for the STM Power Control (PWR) device.
    '''
    REGS = [AReg32('CR1',       0x000, [('LPMS',        0, 2),
                                        ('DBP',         8),
                                        ('VOS',         9, 10),
                                        ('LPR',         14),
                                        ]),
            AReg32('CR2',       0x004, [('PVDE',        0),
                                        ('PLS',         1, 3),
                                        ('PVMEN1',      6),
                                        ('PVMEN2',      7),
                                        ]),
            AReg32('CR3',       0x008, [('EWUP1',       0),
                                        ('EWUP2',       1),
                                        ('EWUP3',       2),
                                        ('EWUP4',       3),
                                        ('EWUP5',       4),
                                        ('RRS',         8),
                                        ('APC',         10),
                                        ('UCPD1_STDBY', 13),
                                        ('UCPD1_DBDIS', 14),
                                        ('EIWUL',       15),
                                        ]),
            AReg32('CR4',       0x00C, [('WP1',         0),
                                        ('WP2',         1),
                                        ('WP3',         2),
                                        ('WP4',         3),
                                        ('WP5',         4),
                                        ('VBE',         8),
                                        ('VBRS',        9),
                                        ]),
            AReg32('SR1',       0x010, [('WUF1',        0),
                                        ('WUF2',        1),
                                        ('WUF3',        2),
                                        ('WUF4',        3),
                                        ('WUF5',        4),
                                        ('SBF',         8),
                                        ('WUFI',        15),
                                        ]),
            AReg32('SR2',       0x014, [('REGLPS',      8),
                                        ('REGLPF',      9),
                                        ('VOSF',        10),
                                        ('PVDO',        11),
                                        ('PVMO1',       14),
                                        ('PVMO2',       15),
                                        ]),
            AReg32('SCR',       0x018, [('CWUF1',       0),
                                        ('CWUF2',       1),
                                        ('CWUF3',       2),
                                        ('CWUF4',       3),
                                        ('CWUF5',       4),
                                        ('CSBF',        8),
                                        ]),
            AReg32('PUCRA',     0x020, [('PU0',         0),
                                        ('PU1',         1),
                                        ('PU2',         2),
                                        ('PU3',         3),
                                        ('PU4',         4),
                                        ('PU5',         5),
                                        ('PU6',         6),
                                        ('PU7',         7),
                                        ('PU8',         8),
                                        ('PU9',         9),
                                        ('PU10',        10),
                                        ('PU11',        11),
                                        ('PU12',        12),
                                        ('PU13',        13),
                                        # ('PU14',        14),
                                        ('PU15',        15),
                                        ]),
            AReg32('PDCRA',     0x024, [('PD0',         0),
                                        ('PD1',         1),
                                        ('PD2',         2),
                                        ('PD3',         3),
                                        ('PD4',         4),
                                        ('PD5',         5),
                                        ('PD6',         6),
                                        ('PD7',         7),
                                        ('PD8',         8),
                                        ('PD9',         9),
                                        ('PD10',        10),
                                        ('PD11',        11),
                                        ('PD12',        12),
                                        # ('PD13',        13),
                                        ('PD14',        14),
                                        # ('PD15',        15),
                                        ]),
            AReg32('PUCRB',     0x028, [('PU0',         0),
                                        ('PU1',         1),
                                        ('PU2',         2),
                                        ('PU3',         3),
                                        ('PU4',         4),
                                        ('PU5',         5),
                                        ('PU6',         6),
                                        ('PU7',         7),
                                        ('PU8',         8),
                                        ('PU9',         9),
                                        ('PU10',        10),
                                        ('PU11',        11),
                                        ('PU12',        12),
                                        ('PU13',        13),
                                        ('PU14',        14),
                                        ('PU15',        15),
                                        ]),
            AReg32('PDCRB',     0x02C, [('PD0',         0),
                                        ('PD1',         1),
                                        ('PD2',         2),
                                        ('PD3',         3),
                                        # ('PD4',         4),
                                        ('PD5',         5),
                                        ('PD6',         6),
                                        ('PD7',         7),
                                        ('PD8',         8),
                                        ('PD9',         9),
                                        ('PD10',        10),
                                        ('PD11',        11),
                                        ('PD12',        12),
                                        ('PD13',        13),
                                        ('PD14',        14),
                                        ('PD15',        15),
                                        ]),
            AReg32('PUCRC',     0x030, [('PU0',         0),
                                        ('PU1',         1),
                                        ('PU2',         2),
                                        ('PU3',         3),
                                        ('PU4',         4),
                                        ('PU5',         5),
                                        ('PU6',         6),
                                        ('PU7',         7),
                                        ('PU8',         8),
                                        ('PU9',         9),
                                        ('PU10',        10),
                                        ('PU11',        11),
                                        ('PU12',        12),
                                        ('PU13',        13),
                                        ('PU14',        14),
                                        ('PU15',        15),
                                        ]),
            AReg32('PDCRC',     0x034, [('PD0',         0),
                                        ('PD1',         1),
                                        ('PD2',         2),
                                        ('PD3',         3),
                                        ('PD4',         4),
                                        ('PD5',         5),
                                        ('PD6',         6),
                                        ('PD7',         7),
                                        ('PD8',         8),
                                        ('PD9',         9),
                                        ('PD10',        10),
                                        ('PD11',        11),
                                        ('PD12',        12),
                                        ('PD13',        13),
                                        ('PD14',        14),
                                        ('PD15',        15),
                                        ]),
            AReg32('PUCRD',     0x038, [('PU0',         0),
                                        ('PU1',         1),
                                        ('PU2',         2),
                                        ('PU3',         3),
                                        ('PU4',         4),
                                        ('PU5',         5),
                                        ('PU6',         6),
                                        ('PU7',         7),
                                        ('PU8',         8),
                                        ('PU9',         9),
                                        ('PU10',        10),
                                        ('PU11',        11),
                                        ('PU12',        12),
                                        ('PU13',        13),
                                        ('PU14',        14),
                                        ('PU15',        15),
                                        ]),
            AReg32('PDCRD',     0x03C, [('PD0',         0),
                                        ('PD1',         1),
                                        ('PD2',         2),
                                        ('PD3',         3),
                                        ('PD4',         4),
                                        ('PD5',         5),
                                        ('PD6',         6),
                                        ('PD7',         7),
                                        ('PD8',         8),
                                        ('PD9',         9),
                                        ('PD10',        10),
                                        ('PD11',        11),
                                        ('PD12',        12),
                                        ('PD13',        13),
                                        ('PD14',        14),
                                        ('PD15',        15),
                                        ]),
            AReg32('PUCRE',     0x040, [('PU0',         0),
                                        ('PU1',         1),
                                        ('PU2',         2),
                                        ('PU3',         3),
                                        ('PU4',         4),
                                        ('PU5',         5),
                                        ('PU6',         6),
                                        ('PU7',         7),
                                        ('PU8',         8),
                                        ('PU9',         9),
                                        ('PU10',        10),
                                        ('PU11',        11),
                                        ('PU12',        12),
                                        ('PU13',        13),
                                        ('PU14',        14),
                                        ('PU15',        15),
                                        ]),
            AReg32('PDCRE',     0x044, [('PD0',         0),
                                        ('PD1',         1),
                                        ('PD2',         2),
                                        ('PD3',         3),
                                        ('PD4',         4),
                                        ('PD5',         5),
                                        ('PD6',         6),
                                        ('PD7',         7),
                                        ('PD8',         8),
                                        ('PD9',         9),
                                        ('PD10',        10),
                                        ('PD11',        11),
                                        ('PD12',        12),
                                        ('PD13',        13),
                                        ('PD14',        14),
                                        ('PD15',        15),
                                        ]),
            AReg32('PUCRF',     0x048, [('PU0',         0),
                                        ('PU1',         1),
                                        ('PU2',         2),
                                        ('PU3',         3),
                                        ('PU4',         4),
                                        ('PU5',         5),
                                        ('PU6',         6),
                                        ('PU7',         7),
                                        ('PU8',         8),
                                        ('PU9',         9),
                                        ('PU10',        10),
                                        ('PU11',        11),
                                        ('PU12',        12),
                                        ('PU13',        13),
                                        ('PU14',        14),
                                        ('PU15',        15),
                                        ]),
            AReg32('PDCRF',     0x04C, [('PD0',         0),
                                        ('PD1',         1),
                                        ('PD2',         2),
                                        ('PD3',         3),
                                        ('PD4',         4),
                                        ('PD5',         5),
                                        ('PD6',         6),
                                        ('PD7',         7),
                                        ('PD8',         8),
                                        ('PD9',         9),
                                        ('PD10',        10),
                                        ('PD11',        11),
                                        ('PD12',        12),
                                        ('PD13',        13),
                                        ('PD14',        14),
                                        ('PD15',        15),
                                        ]),
            AReg32('PUCRG',     0x050, [('PU0',         0),
                                        ('PU1',         1),
                                        ('PU2',         2),
                                        ('PU3',         3),
                                        ('PU4',         4),
                                        ('PU5',         5),
                                        ('PU6',         6),
                                        ('PU7',         7),
                                        ('PU8',         8),
                                        ('PU9',         9),
                                        ('PU10',        10),
                                        # ('PU11',        11),
                                        # ('PU12',        12),
                                        # ('PU13',        13),
                                        # ('PU14',        14),
                                        # ('PU15',        15),
                                        ]),
            AReg32('PDCRG',     0x054, [('PD0',         0),
                                        ('PD1',         1),
                                        ('PD2',         2),
                                        ('PD3',         3),
                                        ('PD4',         4),
                                        ('PD5',         5),
                                        ('PD6',         6),
                                        ('PD7',         7),
                                        ('PD8',         8),
                                        ('PD9',         9),
                                        ('PD10',        10),
                                        # ('PD11',        11),
                                        # ('PD12',        12),
                                        # ('PD13',        13),
                                        # ('PD14',        14),
                                        # ('PD15',        15),
                                        ]),
            AReg32('CR5',       0x080, [('R1MODE',      8),
                                        ]),
            ]

    def __init__(self, target, ap, name, addr, **kwargs):
        super(PWR, self).__init__(target, ap, addr, name, PWR.REGS, **kwargs)

    def unlock_backup_domain(self):
        self._CR1.DBP = 1
        while self._CR1.DBP == 0:
            time.sleep(0.01)
