# Copyright (c) 2020 Phase Advanced Sensor Systems, Inc.
import time
import math

from ..device import Device, Reg32


ENABLE_BITS = {
    # AHB3
    'MDMA'              : (0xD4,  0),
    'DMA2D'             : (0xD4,  4),
    'JPGDEC'            : (0xD4,  5),
    'FILT'              : (0xD4,  8),
    'FMC'               : (0xD4, 12),
    'QSPI'              : (0xD4, 14),
    'SDMMC1'            : (0xD4, 16),
    'DTCM1'             : (0xD4, 28),
    'DTCM2'             : (0xD4, 29),
    'ITCM'              : (0xD4, 30),
    'AXISRAM'           : (0xD4, 31),

    # AHB1
    'DMA1'              : (0xD8,  0),
    'DMA2'              : (0xD8,  1),
    'ADC12'             : (0xD8,  5),
    'ART'               : (0xD8, 14),
    'ETH1MAC'           : (0xD8, 15),
    'ETH1TX'            : (0xD8, 16),
    'ETH1RX'            : (0xD8, 17),
    'USB2OTGHSULPI'     : (0xD8, 18),
    'USB1OTGHS'         : (0xD8, 25),
    'USB1OTGHSULPI'     : (0xD8, 26),
    'USB2OTGHS'         : (0xD8, 27),

    # AHB2
    'DCMI'              : (0xDC,  0),
    'CRYPT'             : (0xDC,  4),
    'HASH'              : (0xDC,  5),
    'RNG'               : (0xDC,  6),
    'SDMMC2'            : (0xDC,  9),
    'SRAM1'             : (0xDC, 29),
    'SRAM2'             : (0xDC, 30),
    'SRAM3'             : (0xDC, 31),

    # AHB4
    'GPIOA'             : (0xE0,  0),
    'GPIOB'             : (0xE0,  1),
    'GPIOC'             : (0xE0,  2),
    'GPIOD'             : (0xE0,  3),
    'GPIOE'             : (0xE0,  4),
    'GPIOF'             : (0xE0,  5),
    'GPIOG'             : (0xE0,  6),
    'GPIOH'             : (0xE0,  7),
    'GPIOI'             : (0xE0,  8),
    'GPIOJ'             : (0xE0,  9),
    'GPIOK'             : (0xE0, 10),
    'CRC'               : (0xE0, 19),
    'BDMA'              : (0xE0, 21),
    'ADC3'              : (0xE0, 24),
    'HSEM'              : (0xE0, 25),
    'BKPRAM'            : (0xE0, 28),

    # APB3
    'LTDC'              : (0xE4,  3),
    'DSI'               : (0xE4,  4),
    'WWDG1'             : (0xE4,  6),

    # APB1L
    'TIM2'              : (0xE8,  0),
    'TIM3'              : (0xE8,  1),
    'TIM4'              : (0xE8,  2),
    'TIM5'              : (0xE8,  3),
    'TIM6'              : (0xE8,  4),
    'TIM7'              : (0xE8,  5),
    'TIM12'             : (0xE8,  6),
    'TIM13'             : (0xE8,  7),
    'TIM14'             : (0xE8,  8),
    'LPTIM1'            : (0xE8,  9),
    'WWDG2'             : (0xE8, 11),
    'SPI2'              : (0xE8, 14),
    'SPI3'              : (0xE8, 15),
    'SPDIFRX'           : (0xE8, 16),
    'USART2'            : (0xE8, 17),
    'USART3'            : (0xE8, 18),
    'UART4'             : (0xE8, 19),
    'UART5'             : (0xE8, 20),
    'I2C1'              : (0xE8, 21),
    'I2C2'              : (0xE8, 22),
    'I2C3'              : (0xE8, 23),
    'CEC'               : (0xE8, 27),
    'DAC1'              : (0xE8, 29),
    'UART7'             : (0xE8, 30),
    'UART8'             : (0xE8, 31),

    # APB1H
    'CRS'               : (0xEC,  1),
    'SWP'               : (0xEC,  2),
    'OPAMP'             : (0xEC,  4),
    'MDIOS'             : (0xEC,  5),
    'FDCAN'             : (0xEC,  8),

    # APB2
    'TIM1'              : (0xF0,  0),
    'TIM8'              : (0xF0,  1),
    'USART1'            : (0xF0,  4),
    'USART6'            : (0xF0,  5),
    'SPI1'              : (0xF0, 12),
    'SPI4'              : (0xF0, 13),
    'TIM15'             : (0xF0, 16),
    'TIM16'             : (0xF0, 17),
    'TIM17'             : (0xF0, 18),
    'SPI5'              : (0xF0, 20),
    'SAI1'              : (0xF0, 22),
    'SAI2'              : (0xF0, 23),
    'SAI3'              : (0xF0, 24),
    'DFSDM1'            : (0xF0, 28),
    'HRTIM'             : (0xF0, 29),

    # APB4
    'SYSCFG'            : (0xF4,  1),
    'LPUART1'           : (0xF4,  3),
    'SPI6'              : (0xF4,  5),
    'I2C4'              : (0xF4,  7),
    'LPTIM2'            : (0xF4,  9),
    'LPTIM3'            : (0xF4, 10),
    'LPTIM4'            : (0xF4, 11),
    'LPTIM5'            : (0xF4, 12),
    'COMP12'            : (0xF4, 14),
    'VREFEN'            : (0xF4, 15),
    'RTCAPB'            : (0xF4, 16),
    'SAI4'              : (0xF4, 21),
    }

# PRE_REV_MAP maps prescaler register bits to a divider value.
# PRE_MAP maps divider values to prescaler register bits.
PRE_REV_MAP = [1, 1, 1, 1, 1, 1, 1, 1, 2, 4, 8, 16, 64, 128, 256, 512]
PRE_MAP = {PRE_REV_MAP[i] : i for i in reversed(range(len(PRE_REV_MAP)))}

# DPRE_REV_MAP maps prescaler register bits to a divider value.
# DPRE_MAP maps divider values to prescaler register bits.
DPRE_REV_MAP = [1, 1, 1, 1, 2, 4, 8, 16]
DPRE_MAP = {DPRE_REV_MAP[i] : i for i in reversed(range(len(DPRE_REV_MAP)))}

# TIM_KER_DIVIDER maps the concatenation of TIMPRE | D2PPREx to a timer kernel
# clock divider.
TIM_KER_DIVIDER = [1, 1, 1, 1, 1, 2, 4, 8, 1, 1, 1, 1, 1, 1, 2, 4]


class RCC(Device):
    '''
    Driver for STM32H745 RCC device.
    '''
    REGS = [Reg32('CR',             0x000, [('HSION',               1),
                                            ('HSIKERON',            1),
                                            ('HSIRDY',              1),
                                            ('HSIDIV',              2),
                                            ('HSIDIVF',             1),
                                            ('',                    1),
                                            ('CSION',               1),
                                            ('CSIRDY',              1),
                                            ('CSIKERON',            1),
                                            ('',                    2),
                                            ('HSI48ON',             1),
                                            ('HSI48RDY',            1),
                                            ('D1CKRDY',             1),
                                            ('D2CKRDY',             1),
                                            ('HSEON',               1),
                                            ('HSERDY',              1),
                                            ('HSEBYP',              1),
                                            ('HSECSSON',            1),
                                            ('',                    4),
                                            ('PLL1ON',              1),
                                            ('PLL1RDY',             1),
                                            ('PLL2ON',              1),
                                            ('PLL2RDY',             1),
                                            ('PLL3ON',              1),
                                            ('PLL3RDY',             1),
                                            ]),
            Reg32('HSICFGR',       0x004,  [('HSICAL',              12),
                                            ('',                    12),
                                            ('HSITRIM',             7),
                                            ]),
            Reg32('CRRCR',         0x008,  [('HSI48CAL',            10),
                                            ]),
            Reg32('CSICFGR',       0x00C,  [('CSICAL',              10),
                                            ('',                    14),
                                            ('CSITRIM',             6),
                                            ]),
            Reg32('CFGR',          0x010,  [('SW',                  3),
                                            ('SWS',                 3),
                                            ('STOPWUCK',            1),
                                            ('STOPKERWUCK',         1),
                                            ('RTCPRE',              6),
                                            ('HRTIMSEL',            1),
                                            ('TIMPRE',              1),
                                            ('',                    2),
                                            ('MCO1PRE',             4),
                                            ('MCO1SEL',             3),
                                            ('MCO2PRE',             4),
                                            ('MCO2SEL',             3),
                                            ]),
            Reg32('D1CFGR',        0x018,  [('HPRE',                4),
                                            ('D1PPRE',              3),
                                            ('',                    1),
                                            ('D1CPRE',              4),
                                            ]),
            Reg32('D2CFGR',        0x01C,  [('',                    4),
                                            ('D2PPRE1',             3),
                                            ('',                    1),
                                            ('D2PPRE2',             3),
                                            ]),
            Reg32('D3CFGR',        0x020,  [('',                    4),
                                            ('D3PPRE',              3),
                                            ]),
            Reg32('PLLCKSELR',     0x028,  [('PLLSRC',              2),
                                            ('',                    2),
                                            ('DIVM1',               6),
                                            ('',                    2),
                                            ('DIVM2',               6),
                                            ('',                    2),
                                            ('DIVM3',               6),
                                            ]),
            Reg32('PLLCFGR',       0x02C,  [('PLL1FRACEN',          1),
                                            ('PLL1VCOSEL',          1),
                                            ('PLL1RGE',             2),
                                            ('PLL2FRACEN',          1),
                                            ('PLL2CVOSEL',          1),
                                            ('PLL2RGE',             2),
                                            ('PLL3FRACEN',          1),
                                            ('PLL3VCOSEL',          1),
                                            ('PLL3RGE',             2),
                                            ('',                    4),
                                            ('DIVP1EN',             1),
                                            ('DIVQ1EN',             1),
                                            ('DIVR1EN',             1),
                                            ('DIVP2EN',             1),
                                            ('DIVQ2EN',             1),
                                            ('DIVR2EN',             1),
                                            ('DIVP3EN',             1),
                                            ('DIVQ3EN',             1),
                                            ('DIVR3EN',             1),
                                            ]),
            Reg32('PLL1DIVR',      0x030,  [('DIVN1',               9),
                                            ('DIVP1',               7),
                                            ('DIVQ1',               7),
                                            ('',                    1),
                                            ('DIVR1',               7),
                                            ]),
            Reg32('PLL1FRACR',     0x034,  [('',                    3),
                                            ('FRACN1',              13),
                                            ]),
            Reg32('PLL2DIVR',      0x038,  [('DIVN2',               9),
                                            ('DIVP2',               7),
                                            ('DIVQ2',               7),
                                            ('',                    1),
                                            ('DIVR2',               7),
                                            ]),
            Reg32('PLL2FRACR',     0x03C,  [('',                    3),
                                            ('FRACN1',              13),
                                            ]),
            Reg32('PLL3DIVR',      0x040,  [('DIVN3',               9),
                                            ('DIVP3',               7),
                                            ('DIVQ3',               7),
                                            ('',                    1),
                                            ('DIVR3',               7),
                                            ]),
            Reg32('PLL3FRACR',     0x044,  [('',                    3),
                                            ('FRACN3',              13),
                                            ]),
            Reg32('D1CCIPR',       0x04C,  [('FMCSEL',              2),
                                            ('',                    2),
                                            ('QSPISEL',             2),
                                            ('',                    2),
                                            ('DSISEL',              1),
                                            ('',                    7),
                                            ('SDMMCSEL',            1),
                                            ('',                    11),
                                            ('CKPERSEL',            2),
                                            ]),
            Reg32('D2CCIP1R',      0x050,  [('SAI1SEL',             3),
                                            ('',                    3),
                                            ('SAI23SEL',            3),
                                            ('',                    3),
                                            ('SPI123SEL',           3),
                                            ('',                    1),
                                            ('SPI45SEL',            3),
                                            ('',                    1),
                                            ('SPDIFSEL',            2),
                                            ('',                    2),
                                            ('DFSDM1SEL',           1),
                                            ('',                    3),
                                            ('FDCANSEL',            2),
                                            ('',                    1),
                                            ('SWPSEL',              1),
                                            ]),
            Reg32('D3CCIPR',       0x058,  [('LPUART1SEL',          3),
                                            ('',                    5),
                                            ('I2C4SEL',             2),
                                            ('LPTIM2SEL',           3),
                                            ('LPTIM345SEL',         3),
                                            ('ADCSEL',              2),
                                            ('',                    3),
                                            ('SAI4ASEL',            3),
                                            ('SAI4BSEL',            3),
                                            ('',                    1),
                                            ('SPI6SEL',             3),
                                            ]),
            Reg32('CIER',          0x060,  [('LSIRDYIE',            1),
                                            ('LSERDYIE',            1),
                                            ('HSIRDYIE',            1),
                                            ('HSERDYIE',            1),
                                            ('CSIRDYIE',            1),
                                            ('HSI48RDYIE',          1),
                                            ('PLL1RDYIE',           1),
                                            ('PLL2RDYIE',           1),
                                            ('PLL3RDYIE',           1),
                                            ('LSECSSIE',            1),
                                            ]),
            Reg32('CIFR',          0x064,  [('LSIRDYF',             1),
                                            ('LSERDYF',             1),
                                            ('HSIRDYF',             1),
                                            ('HSERDYF',             1),
                                            ('CSIRDYF',             1),
                                            ('HSI48RDYF',           1),
                                            ('PLL1RDYF',            1),
                                            ('PLL2RDYF',            1),
                                            ('PLL3RDYF',            1),
                                            ('LSECSSF',             1),
                                            ('HSECSSF',             1),
                                            ]),
            Reg32('CICR',          0x068,  [('LSIRDYC',             1),
                                            ('LSERDYC',             1),
                                            ('HSIRDYC',             1),
                                            ('HSERDYC',             1),
                                            ('CSIRDYC',             1),
                                            ('HSI48RDYC',           1),
                                            ('PLL1RDYC',            1),
                                            ('PLL2RDYC',            1),
                                            ('PLL3RDYC',            1),
                                            ('LSECSSC',             1),
                                            ('HSECSSC',             1),
                                            ]),
            Reg32('BDCR',          0x070,  [('LSEON',               1),
                                            ('LSERDY',              1),
                                            ('LSEBYP',              1),
                                            ('LSEDRV',              2),
                                            ('LSECSSON',            1),
                                            ('LSECSSD',             1),
                                            ('',                    1),
                                            ('RTCSEL',              2),
                                            ('',                    5),
                                            ('RTCEN',               1),
                                            ('BDRST',               1),
                                            ]),
            Reg32('CSR',           0x074,  [('LSION',               1),
                                            ('LSIRDY',              1),
                                            ]),
            Reg32('AHB3RSTR',      0x07C,  [('MDMARST',             1),
                                            ('',                    3),
                                            ('DMA2DRST',            1),
                                            ('JPGDECRST',           1),
                                            ('',                    6),
                                            ('FMCRST',              1),
                                            ('',                    1),
                                            ('QSPIRST',             1),
                                            ('',                    1),
                                            ('SDMMC1RST',           1),
                                            ]),
            Reg32('AHB1RSTR',      0x080,  [('DMA1RST',             1),
                                            ('DMA2RST',             1),
                                            ('',                    3),
                                            ('ADC12RST',            1),
                                            ('',                    8),
                                            ('ARTRST',              1),
                                            ('ETH1MACRST',          1),
                                            ('',                    9),
                                            ('USB1OTGRST',          1),
                                            ('',                    1),
                                            ('USB2OTGRST',          1),
                                            ]),
            Reg32('AHB2RSTR',      0x084,  [('CAMITFRST',           1),
                                            ('',                    3),
                                            ('CRYPTRST',            1),
                                            ('HASHRST',             1),
                                            ('RNGRST',              1),
                                            ('',                    2),
                                            ('SDMMC2RST',           1),
                                            ]),
            Reg32('AHB4RSTR',      0x088,  [('GPIOARST',            1),
                                            ('GPIOBRST',            1),
                                            ('GPIOCRST',            1),
                                            ('GPIODRST',            1),
                                            ('GPIOERST',            1),
                                            ('GPIOFRST',            1),
                                            ('GPIOGRST',            1),
                                            ('GPIOHRST',            1),
                                            ('GPIOIRST',            1),
                                            ('GPIOJRST',            1),
                                            ('GPIOKRST',            1),
                                            ('',                    8),
                                            ('CRCRST',              1),
                                            ('',                    1),
                                            ('BDMARST',             1),
                                            ('',                    2),
                                            ('ADC3RST',             1),
                                            ('HSEMRST',             1),
                                            ]),
            Reg32('APB3RSTR',      0x08C,  [('',                    3),
                                            ('LTDCRST',             1),
                                            ('DSIRST',              1),
                                            ]),
            Reg32('APB1LRSTR',     0x090,  [('TIM2RST',             1),
                                            ('TIM3RST',             1),
                                            ('TIM4RST',             1),
                                            ('TIM5RST',             1),
                                            ('TIM6RST',             1),
                                            ('TIM7RST',             1),
                                            ('TIM12RST',            1),
                                            ('TIM13RST',            1),
                                            ('TIM14RST',            1),
                                            ('LPTIM1RST',           1),
                                            ('',                    4),
                                            ('SPI2RST',             1),
                                            ('SPI3RST',             1),
                                            ('SPDIFRXRST',          1),
                                            ('USART2RST',           1),
                                            ('USART3RST',           1),
                                            ('UART4RST',            1),
                                            ('UART5RST',            1),
                                            ('I2C1RST',             1),
                                            ('I2C2RST',             1),
                                            ('I2C3RST',             1),
                                            ('',                    3),
                                            ('CECRST',              1),
                                            ('',                    1),
                                            ('DAC1RST',             1),
                                            ('UART7RST',            1),
                                            ('UART8RST',            1),
                                            ]),
            Reg32('APB1HRSTR',     0x094,  [('',                    1),
                                            ('CRSRST',              1),
                                            ('SWPRST',              1),
                                            ('',                    1),
                                            ('OPAMPRST',            1),
                                            ('MDIOSRST',            1),
                                            ('',                    2),
                                            ('FDCANRST',            1),
                                            ]),
            Reg32('APB2RSTR',      0x098,  [('TIM1RST',             1),
                                            ('TIM8RST',             1),
                                            ('',                    2),
                                            ('USART1RST',           1),
                                            ('USART6RST',           1),
                                            ('',                    6),
                                            ('SPI1RST',             1),
                                            ('SPI4RST',             1),
                                            ('',                    2),
                                            ('TIM15RST',            1),
                                            ('TIM16RST',            1),
                                            ('TIM17RST',            1),
                                            ('',                    1),
                                            ('SPI5RST',             1),
                                            ('',                    1),
                                            ('SAI1RST',             1),
                                            ('SAI2RST',             1),
                                            ('SAI3RST',             1),
                                            ('',                    3),
                                            ('DFSDM1RST',           1),
                                            ('HRTIMRST',            1),
                                            ]),
            Reg32('APB4RSTR',      0x09C,  [('',                    1),
                                            ('SYSCFGRST',           1),
                                            ('',                    1),
                                            ('LPUART1RST',          1),
                                            ('',                    1),
                                            ('SPI6RST',             1),
                                            ('',                    1),
                                            ('I2C4RST',             1),
                                            ('',                    1),
                                            ('LPTIM2RST',           1),
                                            ('LPTIM3RST',           1),
                                            ('LPTIM4RST',           1),
                                            ('LPTIM5RST',           1),
                                            ('',                    1),
                                            ('COMP12RST',           1),
                                            ('VREFRST',             1),
                                            ('',                    5),
                                            ('SAI4RST',             1),
                                            ]),
            Reg32('GCR',           0x0A0,  [('WW1RSC',              1),
                                            ('WW2RSC',              2),
                                            ('BOOT_C1',             1),
                                            ('BOOT_C2',             1),
                                            ]),
            Reg32('D3AMR',         0x0A8,  [('BDMAAMEN',            1),
                                            ('',                    2),
                                            ('LPUART1AMEN',         1),
                                            ('',                    1),
                                            ('SPI6AMEN',            1),
                                            ('',                    1),
                                            ('I2C4AMEN',            1),
                                            ('',                    1),
                                            ('LPTIM2AMEN',          1),
                                            ('LPTIM3AMEN',          1),
                                            ('LPTIM4AMEN',          1),
                                            ('LPTIM5AMEN',          1),
                                            ('',                    1),
                                            ('COMP12AMEN',          1),
                                            ('VREFAMEN',            1),
                                            ('RTCAMEN',             1),
                                            ('',                    2),
                                            ('CRCAMEN',             1),
                                            ('',                    1),
                                            ('SAI4AMEN',            1),
                                            ('',                    2),
                                            ('ADC3AMEN',            1),
                                            ('',                    3),
                                            ('BKPRAMAMEN',          1),
                                            ('SRAM4AMEN',           1),
                                            ]),

            Reg32('RSR',           0x0D0,  [('',                    16),
                                            ('RMVF',                1),
                                            ('C1RSTF',              1),
                                            ('C2RSTF',              1),
                                            ('D1RSTF',              1),
                                            ('D2RSTF',              1),
                                            ('BORRSTF',             1),
                                            ('PINRSTF',             1),
                                            ('PORRSTF',             1),
                                            ('SFT1RSTF',            1),
                                            ('SFT2RSTF',            1),
                                            ('IWDG1RSTF',           1),
                                            ('IWDG2RSTF',           1),
                                            ('WWDG1RSTF',           1),
                                            ('WWDG2RSTF',           1),
                                            ('LPWR1RSTF',           1),
                                            ('LPWR2RSTF',           1),
                                            ]),
            Reg32('AHB3ENR',       0x0D4,  [('MDMAEN',              1),
                                            ('',                    3),
                                            ('DMA2DEN',             1),
                                            ('JPGDECEN',            1),
                                            ('',                    2),
                                            ('FILTEN',              1),
                                            ('',                    3),
                                            ('FMCEN',               1),
                                            ('',                    1),
                                            ('QSPIEN',              1),
                                            ('',                    1),
                                            ('SDMMC1EN',            1),
                                            ('',                    11),
                                            ('DTCM1EN',             1),
                                            ('DTCM2EN',             1),
                                            ('ITCMEN',              1),
                                            ('AXISRAMEN',           1),
                                            ]),
            Reg32('AHB1ENR',       0x0D8,  [('DMA1EN',              1),
                                            ('DMA2EN',              1),
                                            ('',                    3),
                                            ('ADC12EN',             1),
                                            ('',                    8),
                                            ('ARTEN',               1),
                                            ('ETH1MACEN',           1),
                                            ('ETH1TXEN',            1),
                                            ('ETH1RXEN',            1),
                                            ('USB2OTGHSULPIEN',     1),
                                            ('',                    6),
                                            ('USB1OTGEN',           1),
                                            ('USB1OTFHSULPIEN',     1),
                                            ('USB2OTGEN',           1),
                                            ]),
            Reg32('AHB2ENR',       0x0DC,  [('DCMIEN',              1),
                                            ('',                    3),
                                            ('CRYPTEN',             1),
                                            ('HASHEN',              1),
                                            ('RNGEN',               1),
                                            ('',                    2),
                                            ('SDMMC2EN',            1),
                                            ('',                    19),
                                            ('SRAM1EN',             1),
                                            ('SRAM2EN',             1),
                                            ('SRAM3EN',             1),
                                            ]),
            Reg32('AHB4ENR',       0x0E0,  [('GPIOAEN',             1),
                                            ('GPIOBEN',             1),
                                            ('GPIOCEN',             1),
                                            ('GPIODEN',             1),
                                            ('GPIOEEN',             1),
                                            ('GPIOFEN',             1),
                                            ('GPIOGEN',             1),
                                            ('GPIOHEN',             1),
                                            ('GPIOIEN',             1),
                                            ('GPIOJEN',             1),
                                            ('GPIOKEN',             1),
                                            ('',                    8),
                                            ('CRCEN',               1),
                                            ('',                    1),
                                            ('BDMAEN',              1),
                                            ('',                    2),
                                            ('ADC3EN',              1),
                                            ('HSEMEN',              1),
                                            ('',                    2),
                                            ('BKPRAMEN',            1),
                                            ]),
            Reg32('APB3ENR',       0x0E4,  [('',                    3),
                                            ('LTDCEN',              1),
                                            ('DSIEN',               1),
                                            ('',                    1),
                                            ('WWDG1EN',             1),
                                            ]),
            Reg32('APB1LENR',      0x0E8,  [('TIM2EN',              1),
                                            ('TIM3EN',              1),
                                            ('TIM4EN',              1),
                                            ('TIM5EN',              1),
                                            ('TIM6EN',              1),
                                            ('TIM7EN',              1),
                                            ('TIM12EN',             1),
                                            ('TIM13EN',             1),
                                            ('TIM14EN',             1),
                                            ('LPTIM1EN',            1),
                                            ('',                    1),
                                            ('WWDG2EN',             1),
                                            ('',                    2),
                                            ('SPI2EN',              1),
                                            ('SPI3EN',              1),
                                            ('SPDIFRXEN',           1),
                                            ('USART2EN',            1),
                                            ('USART3EN',            1),
                                            ('UART4EN',             1),
                                            ('UART5EN',             1),
                                            ('I2C1EN',              1),
                                            ('I2C2EN',              1),
                                            ('I2C3EN',              1),
                                            ('',                    3),
                                            ('CECEN',               1),
                                            ('',                    1),
                                            ('DAC1EN',              1),
                                            ('UART7EN',             1),
                                            ('UART8EN',             1),
                                            ]),
            Reg32('APB1HENR',      0x0EC,  [('',                    1),
                                            ('CRSEN',               1),
                                            ('SWPEN',               1),
                                            ('',                    1),
                                            ('OPAMPEN',             1),
                                            ('MDIOSEN',             1),
                                            ('',                    2),
                                            ('FDCANEN',             1),
                                            ]),
            Reg32('APB2ENR',       0x0F0,  [('TIM1EN',              1),
                                            ('TIM8EN',              1),
                                            ('',                    2),
                                            ('USART1EN',            1),
                                            ('USART6EN',            1),
                                            ('',                    6),
                                            ('SPI1EN',              1),
                                            ('SPI4EN',              1),
                                            ('',                    2),
                                            ('TIM15EN',             1),
                                            ('TIM16EN',             1),
                                            ('TIM17EN',             1),
                                            ('',                    1),
                                            ('SPI5EN',              1),
                                            ('',                    1),
                                            ('SAI1EN',              1),
                                            ('SAI2EN',              1),
                                            ('SAI3EN',              1),
                                            ('',                    3),
                                            ('DFSDM1EN',            1),
                                            ('HRTIMEN',             1),
                                            ]),
            Reg32('APB4ENR',       0x0F4,  [('',                    1),
                                            ('SYSCFGEN',            1),
                                            ('',                    1),
                                            ('LPUART1EN',           1),
                                            ('',                    1),
                                            ('SPI6EN',              1),
                                            ('',                    1),
                                            ('I2C4EN',              1),
                                            ('',                    1),
                                            ('LPTIM2EN',            1),
                                            ('LPTIM3EN',            1),
                                            ('LPTIM4EN',            1),
                                            ('LPTIM5EN',            1),
                                            ('',                    1),
                                            ('COMP12EN',            1),
                                            ('VREFEN',              1),
                                            ('RTCAPBEN',            1),
                                            ('',                    4),
                                            ('SAI4EN',              1),
                                            ]),
            Reg32('AHB3LPENR',     0x0FC,  [('MDMALPEN',            1),
                                            ('',                    3),
                                            ('DMA2DLPEN',           1),
                                            ('JPGDECLPEN',          1),
                                            ('',                    2),
                                            ('FILTLPEN',            1),
                                            ('',                    3),
                                            ('FMCLPEN',             1),
                                            ('',                    1),
                                            ('QSPILPEN',            1),
                                            ('',                    1),
                                            ('SDMMC1LPEN',          1),
                                            ('',                    11),
                                            ('DTCM1LPEN',           1),
                                            ('DTCM2LPEN',           1),
                                            ('ITCMLPEN',            1),
                                            ('AXISRAMLPEN',         1),
                                            ]),
            Reg32('AHB1LPENR',     0x100,  [('DMA1LPEN',            1),
                                            ('DMA2LPEN',            1),
                                            ('',                    3),
                                            ('ADC12LPEN',           1),
                                            ('',                    8),
                                            ('ARTLPEN',             1),
                                            ('ETH1MACLPEN',         1),
                                            ('ETH1TXLPEN',          1),
                                            ('ETH1RXLPEN',          1),
                                            ('',                    7),
                                            ('USB1OTGLPEN',         1),
                                            ('USB1OTGHSULPILPEN',   1),
                                            ('USB2OTGLPEN',         1),
                                            ('USB2OTGHSULPILPEN',   1),
                                            ]),
            Reg32('AHB2LPENR',     0x104,  [('CAMITFLPEN',          1),
                                            ('',                    3),
                                            ('CRYPTLPEN',           1),
                                            ('HASHLPEN',            1),
                                            ('RNGLPEN',             1),
                                            ('',                    2),
                                            ('SDMMC2LPEN',          1),
                                            ('',                    19),
                                            ('SRAM1LPEN',           1),
                                            ('SRAM2LPEN',           1),
                                            ('SRAM3LPEN',           1),
                                            ]),
            Reg32('AHB4LPENR',     0x108,  [('GPIOALPEN',           1),
                                            ('GPIOBLPEN',           1),
                                            ('GPIOCLPEN',           1),
                                            ('GPIODLPEN',           1),
                                            ('GPIOELPEN',           1),
                                            ('GPIOFLPEN',           1),
                                            ('GPIOGLPEN',           1),
                                            ('GPIOHLPEN',           1),
                                            ('GPIOILPEN',           1),
                                            ('GPIOJLPEN',           1),
                                            ('GPIOKLPEN',           1),
                                            ('',                    8),
                                            ('CRCLPEN',             1),
                                            ('',                    1),
                                            ('BDMALPEN',            1),
                                            ('',                    2),
                                            ('ADC3LPEN',            1),
                                            ('',                    3),
                                            ('BKPRAMLPEN',          1),
                                            ('SRAM4LPEN',           1),
                                            ]),
            Reg32('APB3LPENR',     0x10C,  [('',                    3),
                                            ('LTDCLPEN',            1),
                                            ('DSILPEN',             1),
                                            ('',                    1),
                                            ('WWDG1LPEN',           1),
                                            ]),
            Reg32('APB1LLPENR',    0x110,  [('TIM2LPEN',            1),
                                            ('TIM3LPEN',            1),
                                            ('TIM4LPEN',            1),
                                            ('TIM5LPEN',            1),
                                            ('TIM6LPEN',            1),
                                            ('TIM7LPEN',            1),
                                            ('TIM12LPEN',           1),
                                            ('TIM13LPEN',           1),
                                            ('TIM14LPEN',           1),
                                            ('LPTIM1LPEN',          1),
                                            ('',                    1),
                                            ('WWDG2LPEN',           1),
                                            ('',                    2),
                                            ('SPI2LPEN',            1),
                                            ('SPI3LPEN',            1),
                                            ('SPDIFRXLPEN',         1),
                                            ('USART2LPEN',          1),
                                            ('USART3LPEN',          1),
                                            ('UART4LPEN',           1),
                                            ('UART5LPEN',           1),
                                            ('I2C1LPEN',            1),
                                            ('I2C2LPEN',            1),
                                            ('I2C3LPEN',            1),
                                            ('',                    3),
                                            ('CECLPEN',             1),
                                            ('',                    1),
                                            ('DAC1LPEN',            1),
                                            ('UART7LPEN',           1),
                                            ('UART8LPEN',           1),
                                            ]),
            Reg32('APB1HLPENR',    0x114,  [('',                    1),
                                            ('CRSLPEN',             1),
                                            ('SWPLPEN',             1),
                                            ('',                    1),
                                            ('OPAMPLPEN',           1),
                                            ('MDIOSLPEN',           1),
                                            ('',                    2),
                                            ('FDCANLPEN',           1),
                                            ]),
            Reg32('APB2LPENR',     0x118,  [('TIM1LPEN',            1),
                                            ('TIM8LPEN',            1),
                                            ('',                    2),
                                            ('USART1LPEN',          1),
                                            ('USART6LPEN',          1),
                                            ('',                    6),
                                            ('SPI1LPEN',            1),
                                            ('SPI4LPEN',            1),
                                            ('',                    2),
                                            ('TIM15LPEN',           1),
                                            ('TIM16LPEN',           1),
                                            ('TIM17LPEN',           1),
                                            ('',                    1),
                                            ('SPI5LPEN',            1),
                                            ('',                    1),
                                            ('SAI1LPEN',            1),
                                            ('SAI2LPEN',            1),
                                            ('SAI3LPEN',            1),
                                            ('',                    3),
                                            ('DFSDM1LPEN',          1),
                                            ('HRTIMLPEN',           1),
                                            ]),
            Reg32('APB4LPENR',     0x11C,  [('',                    1),
                                            ('SYSCFGLPEN',          1),
                                            ('',                    1),
                                            ('LPUART1LPEN',         1),
                                            ('',                    1),
                                            ('SPI6LPEN',            1),
                                            ('',                    1),
                                            ('I2C4LPEN',            1),
                                            ('',                    1),
                                            ('LPTIM2LPEN',          1),
                                            ('LPTIM3LPEN',          1),
                                            ('LPTIM4LPEN',          1),
                                            ('LPTIM5LPEN',          1),
                                            ('',                    1),
                                            ('COMP12LPEN',          1),
                                            ('VREFLPEN',            1),
                                            ('RTCAPBLPEN',          1),
                                            ('',                    4),
                                            ('SAI4LPEN',            1),
                                            ]),
            Reg32('C1_RSR',        0x130,  [('',                    16),
                                            ('RMVF',                1),
                                            ('C1RSTF',              1),
                                            ('C2RSTF',              1),
                                            ('D1RSTF',              1),
                                            ('D2RSTF',              1),
                                            ('BORRSTF',             1),
                                            ('PINRSTF',             1),
                                            ('PORRSTF',             1),
                                            ('SFT1RSTF',            1),
                                            ('SFT2RSTF',            1),
                                            ('IWDG1RSTF',           1),
                                            ('IWDG2RSTF',           1),
                                            ('WWDG1RSTF',           1),
                                            ('WWDG2RSTF',           1),
                                            ('LPWR1RSTF',           1),
                                            ('LPWR2RSTF',           1),
                                            ]),
            Reg32('C1_AHB3ENR',    0x134,  [('MDMAEN',              1),
                                            ('',                    3),
                                            ('DMA2DEN',             1),
                                            ('JPGDECEN',            1),
                                            ('',                    6),
                                            ('FMCEN',               1),
                                            ('',                    1),
                                            ('QSPIEN',              1),
                                            ('',                    1),
                                            ('SDMMC1EN',            1),
                                            ]),
            Reg32('C1_AHB1ENR',    0x138,  [('DMA1EN',              1),
                                            ('DMA2EN',              1),
                                            ('',                    3),
                                            ('ADC12EN',             1),
                                            ('',                    8),
                                            ('ARTEN',               1),
                                            ('ETH1MACEN',           1),
                                            ('ETH1TXEN',            1),
                                            ('ETH1RXEN',            1),
                                            ('USB2OTGHSULPIEN',     1),
                                            ('',                    6),
                                            ('USB1OTGEN',           1),
                                            ('USB1OTFHSULPIEN',     1),
                                            ('USB2OTGEN',           1),
                                            ]),
            Reg32('C1_AHB2ENR',    0x13C,  [('DCMIEN',              1),
                                            ('',                    3),
                                            ('CRYPTEN',             1),
                                            ('HASHEN',              1),
                                            ('RNGEN',               1),
                                            ('',                    2),
                                            ('SDMMC2EN',            1),
                                            ('',                    19),
                                            ('SRAM1EN',             1),
                                            ('SRAM2EN',             1),
                                            ('SRAM3EN',             1),
                                            ]),
            Reg32('C1_AHB4ENR',    0x140,  [('GPIOAEN',             1),
                                            ('GPIOBEN',             1),
                                            ('GPIOCEN',             1),
                                            ('GPIODEN',             1),
                                            ('GPIOEEN',             1),
                                            ('GPIOFEN',             1),
                                            ('GPIOGEN',             1),
                                            ('GPIOHEN',             1),
                                            ('GPIOIEN',             1),
                                            ('GPIOJEN',             1),
                                            ('GPIOKEN',             1),
                                            ('',                    8),
                                            ('CRCEN',               1),
                                            ('',                    1),
                                            ('BDMAEN',              1),
                                            ('',                    2),
                                            ('ADC3EN',              1),
                                            ('HSEMEN',              1),
                                            ('',                    2),
                                            ('BKPRAMEN',            1),
                                            ]),
            Reg32('C1_APB3ENR',    0x144,  [('',                    3),
                                            ('LTDCEN',              1),
                                            ('DSIEN',               1),
                                            ('',                    1),
                                            ('WWDG1EN',             1),
                                            ]),
            Reg32('C1_APB1LENR',   0x148,  [('TIM2EN',              1),
                                            ('TIM3EN',              1),
                                            ('TIM4EN',              1),
                                            ('TIM5EN',              1),
                                            ('TIM6EN',              1),
                                            ('TIM7EN',              1),
                                            ('TIM12EN',             1),
                                            ('TIM13EN',             1),
                                            ('TIM14EN',             1),
                                            ('LPTIM1EN',            1),
                                            ('',                    1),
                                            ('WWDG2EN',             1),
                                            ('',                    2),
                                            ('SPI2EN',              1),
                                            ('SPI3EN',              1),
                                            ('SPDIFRXEN',           1),
                                            ('USART2EN',            1),
                                            ('USART3EN',            1),
                                            ('UART4EN',             1),
                                            ('UART5EN',             1),
                                            ('I2C1EN',              1),
                                            ('I2C2EN',              1),
                                            ('I2C3EN',              1),
                                            ('',                    3),
                                            ('CECEN',               1),
                                            ('',                    1),
                                            ('DAC1EN',              1),
                                            ('UART7EN',             1),
                                            ('UART8EN',             1),
                                            ]),
            Reg32('C1_APB1HENR',   0x14C,  [('',                    1),
                                            ('CRSEN',               1),
                                            ('SWPEN',               1),
                                            ('',                    1),
                                            ('OPAMPEN',             1),
                                            ('MDIOSEN',             1),
                                            ('',                    2),
                                            ('FDCANEN',             1),
                                            ]),
            Reg32('C1_APB2ENR',    0x150,  [('TIM1EN',              1),
                                            ('TIM8EN',              1),
                                            ('',                    2),
                                            ('USART1EN',            1),
                                            ('USART6EN',            1),
                                            ('',                    6),
                                            ('SPI1EN',              1),
                                            ('SPI4EN',              1),
                                            ('',                    2),
                                            ('TIM15EN',             1),
                                            ('TIM16EN',             1),
                                            ('TIM17EN',             1),
                                            ('',                    1),
                                            ('SPI5EN',              1),
                                            ('',                    1),
                                            ('SAI1EN',              1),
                                            ('SAI2EN',              1),
                                            ('SAI3EN',              1),
                                            ('',                    3),
                                            ('DFSDM1EN',            1),
                                            ('HRTIMEN',             1),
                                            ]),
            Reg32('C1_APB4ENR',    0x154,  [('',                    1),
                                            ('SYSCFGEN',            1),
                                            ('',                    1),
                                            ('LPUART1EN',           1),
                                            ('',                    1),
                                            ('SPI6EN',              1),
                                            ('',                    1),
                                            ('I2C4EN',              1),
                                            ('',                    1),
                                            ('LPTIM2EN',            1),
                                            ('LPTIM3EN',            1),
                                            ('LPTIM4EN',            1),
                                            ('LPTIM5EN',            1),
                                            ('',                    1),
                                            ('COMP12EN',            1),
                                            ('VREFEN',              1),
                                            ('RTCAPBEN',            1),
                                            ('',                    4),
                                            ('SAI4EN',              1),
                                            ]),
            Reg32('C1_AHB3LPENR',  0x15C,  [('MDMALPEN',            1),
                                            ('',                    3),
                                            ('DMA2DLPEN',           1),
                                            ('JPGDECLPEN',          1),
                                            ('',                    2),
                                            ('FILTLPEN',            1),
                                            ('',                    3),
                                            ('FMCLPEN',             1),
                                            ('',                    1),
                                            ('QSPILPEN',            1),
                                            ('',                    1),
                                            ('SDMMC1LPEN',          1),
                                            ('',                    11),
                                            ('DTCM1LPEN',           1),
                                            ('DTCM2LPEN',           1),
                                            ('ITCMLPEN',            1),
                                            ('AXISRAMLPEN',         1),
                                            ]),
            Reg32('C1_AHB1LPENR',  0x160,  [('DMA1LPEN',            1),
                                            ('DMA2LPEN',            1),
                                            ('',                    3),
                                            ('ADC12LPEN',           1),
                                            ('',                    8),
                                            ('ARTLPEN',             1),
                                            ('ETH1MACLPEN',         1),
                                            ('ETH1TXLPEN',          1),
                                            ('ETH1RXLPEN',          1),
                                            ('',                    8),
                                            ('USB1OTGLPEN',         1),
                                            ('USB1OTGHSULPILPEN',   1),
                                            ('USB2OTGLPEN',         1),
                                            ('USB2OTGHSULPILPEN',   1),
                                            ]),
            Reg32('C1_AHB2LPENR',  0x164,  [('CAMITFLPEN',          1),
                                            ('',                    3),
                                            ('CRYPTLPEN',           1),
                                            ('HASHLPEN',            1),
                                            ('RNGLPEN',             1),
                                            ('',                    2),
                                            ('SDMMC2LPEN',          1),
                                            ('',                    19),
                                            ('SRAM1LPEN',           1),
                                            ('SRAM2LPEN',           1),
                                            ('SRAM3LPEN',           1),
                                            ]),
            Reg32('C1_AHB4LPENR',  0x168,  [('GPIOALPEN',           1),
                                            ('GPIOBLPEN',           1),
                                            ('GPIOCLPEN',           1),
                                            ('GPIODLPEN',           1),
                                            ('GPIOELPEN',           1),
                                            ('GPIOFLPEN',           1),
                                            ('GPIOGLPEN',           1),
                                            ('GPIOHLPEN',           1),
                                            ('GPIOILPEN',           1),
                                            ('GPIOJLPEN',           1),
                                            ('GPIOKLPEN',           1),
                                            ('',                    8),
                                            ('CRCLPEN',             1),
                                            ('',                    1),
                                            ('BDMALPEN',            1),
                                            ('',                    2),
                                            ('ADC3LPEN',            1),
                                            ('',                    3),
                                            ('BKPRAMLPEN',          1),
                                            ('SRAM4LPEN',           1),
                                            ]),
            Reg32('C1_APB3LPENR',  0x16C,  [('',                    3),
                                            ('LTDCLPEN',            1),
                                            ('DSILPEN',             1),
                                            ('',                    1),
                                            ('WWDG1LPEN',           1),
                                            ]),
            Reg32('C1_APB1LLPENR', 0x170,  [('TIM2LPEN',            1),
                                            ('TIM3LPEN',            1),
                                            ('TIM4LPEN',            1),
                                            ('TIM5LPEN',            1),
                                            ('TIM6LPEN',            1),
                                            ('TIM7LPEN',            1),
                                            ('TIM12LPEN',           1),
                                            ('TIM13LPEN',           1),
                                            ('TIM14LPEN',           1),
                                            ('LPTIM1LPEN',          1),
                                            ('',                    1),
                                            ('WWDG2LPEN',           1),
                                            ('',                    2),
                                            ('SPI2LPEN',            1),
                                            ('SPI3LPEN',            1),
                                            ('SPDIFRXLPEN',         1),
                                            ('USART2LPEN',          1),
                                            ('USART3LPEN',          1),
                                            ('UART4LPEN',           1),
                                            ('UART5LPEN',           1),
                                            ('I2C1LPEN',            1),
                                            ('I2C2LPEN',            1),
                                            ('I2C3LPEN',            1),
                                            ('',                    3),
                                            ('CECLPEN',             1),
                                            ('',                    1),
                                            ('DAC1LPEN',            1),
                                            ('UART7LPEN',           1),
                                            ('UART8LPEN',           1),
                                            ]),
            Reg32('C1_APB1HLPENR', 0x174,  [('',                    1),
                                            ('CRSLPEN',             1),
                                            ('SWPLPEN',             1),
                                            ('',                    1),
                                            ('OPAMPLPEN',           1),
                                            ('MDIOSLPEN',           1),
                                            ('',                    2),
                                            ('FDCANLPEN',           1),
                                            ]),
            Reg32('C1_APB2LPENR',  0x178,  [('TIM1LPEN',            1),
                                            ('TIM8LPEN',            1),
                                            ('',                    2),
                                            ('USART1LPEN',          1),
                                            ('USART6LPEN',          1),
                                            ('',                    6),
                                            ('SPI1LPEN',            1),
                                            ('SPI4LPEN',            1),
                                            ('',                    2),
                                            ('TIM15LPEN',           1),
                                            ('TIM16LPEN',           1),
                                            ('TIM17LPEN',           1),
                                            ('',                    1),
                                            ('SPI5LPEN',            1),
                                            ('',                    1),
                                            ('SAI1LPEN',            1),
                                            ('SAI2LPEN',            1),
                                            ('SAI3LPEN',            1),
                                            ('',                    3),
                                            ('DFSDM1LPEN',          1),
                                            ('HRTIMLPEN',           1),
                                            ]),
            Reg32('C1_APB4LPENR',  0x17C,  [('',                    1),
                                            ('SYSCFGLPEN',          1),
                                            ('',                    1),
                                            ('LPUART1LPEN',         1),
                                            ('',                    1),
                                            ('SPI6LPEN',            1),
                                            ('',                    1),
                                            ('I2C4LPEN',            1),
                                            ('',                    1),
                                            ('LPTIM2LPEN',          1),
                                            ('LPTIM3LPEN',          1),
                                            ('LPTIM4LPEN',          1),
                                            ('LPTIM5LPEN',          1),
                                            ('',                    1),
                                            ('COMP12LPEN',          1),
                                            ('VREFLPEN',            1),
                                            ('RTCAPBLPEN',          1),
                                            ('',                    4),
                                            ('SAI4LPEN',            1),
                                            ]),
            Reg32('C2_RSR',        0x190,  [('',                    16),
                                            ('RMVF',                1),
                                            ('C1RSTF',              1),
                                            ('C2RSTF',              1),
                                            ('D1RSTF',              1),
                                            ('D2RSTF',              1),
                                            ('BORRSTF',             1),
                                            ('PINRSTF',             1),
                                            ('PORRSTF',             1),
                                            ('SFT1RSTF',            1),
                                            ('SFT2RSTF',            1),
                                            ('IWDG1RSTF',           1),
                                            ('IWDG2RSTF',           1),
                                            ('WWDG1RSTF',           1),
                                            ('WWDG2RSTF',           1),
                                            ('LPWR1RSTF',           1),
                                            ('LPWR2RSTF',           1),
                                            ]),
            Reg32('C2_AHB3ENR',    0x194,  [('MDMAEN',              1),
                                            ('',                    3),
                                            ('DMA2DEN',             1),
                                            ('JPGDECEN',            1),
                                            ('',                    2),
                                            ('FILTEN',              1),
                                            ('',                    3),
                                            ('FMCEN',               1),
                                            ('',                    1),
                                            ('QSPIEN',              1),
                                            ('',                    1),
                                            ('SDMMC1EN',            1),
                                            ('',                    11),
                                            ('DTCM1EN',             1),
                                            ('DTCM2EN',             1),
                                            ('ITCMEN',              1),
                                            ('AXISRAMEN',           1),
                                            ]),
            Reg32('C2_AHB1ENR',    0x198,  [('DMA1EN',              1),
                                            ('DMA2EN',              1),
                                            ('',                    3),
                                            ('ADC12EN',             1),
                                            ('',                    8),
                                            ('ARTEN',               1),
                                            ('ETH1MACEN',           1),
                                            ('ETH1TXEN',            1),
                                            ('ETH1RXEN',            1),
                                            ('USB2OTGHSULPIEN',     1),
                                            ('',                    6),
                                            ('USB1OTGEN',           1),
                                            ('USB1OTFHSULPIEN',     1),
                                            ('USB2OTGEN',           1),
                                            ]),
            Reg32('C2_AHB2ENR',    0x19C,  [('DCMIEN',              1),
                                            ('',                    3),
                                            ('CRYPTEN',             1),
                                            ('HASHEN',              1),
                                            ('RNGEN',               1),
                                            ('',                    2),
                                            ('SDMMC2EN',            1),
                                            ]),
            Reg32('C2_AHB4ENR',    0x1A0,  [('GPIOAEN',             1),
                                            ('GPIOBEN',             1),
                                            ('GPIOCEN',             1),
                                            ('GPIODEN',             1),
                                            ('GPIOEEN',             1),
                                            ('GPIOFEN',             1),
                                            ('GPIOGEN',             1),
                                            ('GPIOHEN',             1),
                                            ('GPIOIEN',             1),
                                            ('GPIOJEN',             1),
                                            ('GPIOKEN',             1),
                                            ('',                    8),
                                            ('CRCEN',               1),
                                            ('',                    1),
                                            ('BDMAEN',              1),
                                            ('',                    2),
                                            ('ADC3EN',              1),
                                            ('HSEMEN',              1),
                                            ('',                    2),
                                            ('BKPRAMEN',            1),
                                            ]),
            Reg32('C2_APB3ENR',    0x1A4,  [('',                    3),
                                            ('LTDCEN',              1),
                                            ('DSIEN',               1),
                                            ('',                    1),
                                            ('WWDG1EN',             1),
                                            ]),
            Reg32('C2_APB1LENR',   0x1A8,  [('TIM2EN',              1),
                                            ('TIM3EN',              1),
                                            ('TIM4EN',              1),
                                            ('TIM5EN',              1),
                                            ('TIM6EN',              1),
                                            ('TIM7EN',              1),
                                            ('TIM12EN',             1),
                                            ('TIM13EN',             1),
                                            ('TIM14EN',             1),
                                            ('LPTIM1EN',            1),
                                            ('',                    1),
                                            ('WWDG2EN',             1),
                                            ('',                    2),
                                            ('SPI2EN',              1),
                                            ('SPI3EN',              1),
                                            ('SPDIFRXEN',           1),
                                            ('USART2EN',            1),
                                            ('USART3EN',            1),
                                            ('UART4EN',             1),
                                            ('UART5EN',             1),
                                            ('I2C1EN',              1),
                                            ('I2C2EN',              1),
                                            ('I2C3EN',              1),
                                            ('',                    3),
                                            ('CECEN',               1),
                                            ('',                    1),
                                            ('DAC1EN',              1),
                                            ('UART7EN',             1),
                                            ('UART8EN',             1),
                                            ]),
            Reg32('C2_APB1HENR',   0x1AC,  [('',                    1),
                                            ('CRSEN',               1),
                                            ('SWPEN',               1),
                                            ('',                    1),
                                            ('OPAMPEN',             1),
                                            ('MDIOSEN',             1),
                                            ('',                    2),
                                            ('FDCANEN',             1),
                                            ]),
            Reg32('C2_APB2ENR',    0x1B0,  [('TIM1EN',              1),
                                            ('TIM8EN',              1),
                                            ('',                    2),
                                            ('USART1EN',            1),
                                            ('USART6EN',            1),
                                            ('',                    6),
                                            ('SPI1EN',              1),
                                            ('SPI4EN',              1),
                                            ('',                    2),
                                            ('TIM15EN',             1),
                                            ('TIM16EN',             1),
                                            ('TIM17EN',             1),
                                            ('',                    1),
                                            ('SPI5EN',              1),
                                            ('',                    1),
                                            ('SAI1EN',              1),
                                            ('SAI2EN',              1),
                                            ('SAI3EN',              1),
                                            ('',                    3),
                                            ('DFSDM1EN',            1),
                                            ('HRTIMEN',             1),
                                            ]),
            Reg32('C2_APB4ENR',    0x1B4,  [('',                    1),
                                            ('SYSCFGEN',            1),
                                            ('',                    1),
                                            ('LPUART1EN',           1),
                                            ('',                    1),
                                            ('SPI6EN',              1),
                                            ('',                    1),
                                            ('I2C4EN',              1),
                                            ('',                    1),
                                            ('LPTIM2EN',            1),
                                            ('LPTIM3EN',            1),
                                            ('LPTIM4EN',            1),
                                            ('LPTIM5EN',            1),
                                            ('',                    1),
                                            ('COMP12EN',            1),
                                            ('VREFEN',              1),
                                            ('RTCAPBEN',            1),
                                            ('',                    4),
                                            ('SAI4EN',              1),
                                            ]),
            Reg32('C2_AHB3LPENR',  0x1BC,  [('MDMALPEN',            1),
                                            ('',                    3),
                                            ('DMA2DLPEN',           1),
                                            ('JPGDECLPEN',          1),
                                            ('',                    2),
                                            ('FILTLPEN',            1),
                                            ('',                    3),
                                            ('FMCLPEN',             1),
                                            ('',                    1),
                                            ('QSPILPEN',            1),
                                            ('',                    1),
                                            ('SDMMC1LPEN',          1),
                                            ('',                    11),
                                            ('DTCM1LPEN',           1),
                                            ('DTCM2LPEN',           1),
                                            ('ITCMLPEN',            1),
                                            ('AXISRAMLPEN',         1),
                                            ]),
            Reg32('C2_AHB1LPENR',  0x1C0,  [('DMA1LPEN',            1),
                                            ('DMA2LPEN',            1),
                                            ('',                    3),
                                            ('ADC12LPEN',           1),
                                            ('',                    8),
                                            ('ARTLPEN',             1),
                                            ('ETH1MACLPEN',         1),
                                            ('ETH1TXLPEN',          1),
                                            ('ETH1RXLPEN',          1),
                                            ('',                    8),
                                            ('USB1OTGLPEN',         1),
                                            ('USB1OTGHSULPILPEN',   1),
                                            ('USB2OTGLPEN',         1),
                                            ('USB2OTGHSULPILPEN',   1),
                                            ]),
            Reg32('C2_AHB2LPENR',  0x1C4,  [('CAMITFLPEN',          1),
                                            ('',                    3),
                                            ('CRYPTLPEN',           1),
                                            ('HASHLPEN',            1),
                                            ('RNGLPEN',             1),
                                            ('',                    2),
                                            ('SDMMC2LPEN',          1),
                                            ('',                    19),
                                            ('SRAM1LPEN',           1),
                                            ('SRAM2LPEN',           1),
                                            ('SRAM3LPEN',           1),
                                            ]),
            Reg32('C2_AHB4LPENR',  0x1C8,  [('GPIOALPEN',           1),
                                            ('GPIOBLPEN',           1),
                                            ('GPIOCLPEN',           1),
                                            ('GPIODLPEN',           1),
                                            ('GPIOELPEN',           1),
                                            ('GPIOFLPEN',           1),
                                            ('GPIOGLPEN',           1),
                                            ('GPIOHLPEN',           1),
                                            ('GPIOILPEN',           1),
                                            ('GPIOJLPEN',           1),
                                            ('GPIOKLPEN',           1),
                                            ('',                    8),
                                            ('CRCLPEN',             1),
                                            ('',                    1),
                                            ('BDMALPEN',            1),
                                            ('',                    2),
                                            ('ADC3LPEN',            1),
                                            ('',                    3),
                                            ('BKPRAMLPEN',          1),
                                            ('SRAM4LPEN',           1),
                                            ]),
            Reg32('C2_APB3LPENR',  0x1CC,  [('',                    3),
                                            ('LTDCLPEN',            1),
                                            ('DSILPEN',             1),
                                            ('',                    1),
                                            ('WWDG1LPEN',           1),
                                            ]),
            Reg32('C2_APB1LLPENR', 0x1D0,  [('TIM2LPEN',            1),
                                            ('TIM3LPEN',            1),
                                            ('TIM4LPEN',            1),
                                            ('TIM5LPEN',            1),
                                            ('TIM6LPEN',            1),
                                            ('TIM7LPEN',            1),
                                            ('TIM12LPEN',           1),
                                            ('TIM13LPEN',           1),
                                            ('TIM14LPEN',           1),
                                            ('LPTIM1LPEN',          1),
                                            ('',                    1),
                                            ('WWDG2LPEN',           1),
                                            ('',                    2),
                                            ('SPI2LPEN',            1),
                                            ('SPI3LPEN',            1),
                                            ('SPDIFRXLPEN',         1),
                                            ('USART2LPEN',          1),
                                            ('USART3LPEN',          1),
                                            ('UART4LPEN',           1),
                                            ('UART5LPEN',           1),
                                            ('I2C1LPEN',            1),
                                            ('I2C2LPEN',            1),
                                            ('I2C3LPEN',            1),
                                            ('',                    3),
                                            ('CECLPEN',             1),
                                            ('',                    1),
                                            ('DAC1LPEN',            1),
                                            ('UART7LPEN',           1),
                                            ('UART8LPEN',           1),
                                            ]),
            Reg32('C2_APB1HLPENR', 0x1D4,  [('',                    1),
                                            ('CRSLPEN',             1),
                                            ('SWPLPEN',             1),
                                            ('',                    1),
                                            ('OPAMPLPEN',           1),
                                            ('MDIOSLPEN',           1),
                                            ('',                    2),
                                            ('FDCANLPEN',           1),
                                            ]),
            Reg32('C2_APB2LPENR',  0x1D8,  [('TIM1LPEN',            1),
                                            ('TIM8LPEN',            1),
                                            ('',                    2),
                                            ('USART1LPEN',          1),
                                            ('USART6LPEN',          1),
                                            ('',                    6),
                                            ('SPI1LPEN',            1),
                                            ('SPI4LPEN',            1),
                                            ('',                    2),
                                            ('TIM15LPEN',           1),
                                            ('TIM16LPEN',           1),
                                            ('TIM17LPEN',           1),
                                            ('',                    1),
                                            ('SPI5LPEN',            1),
                                            ('',                    1),
                                            ('SAI1LPEN',            1),
                                            ('SAI2LPEN',            1),
                                            ('SAI3LPEN',            1),
                                            ('',                    3),
                                            ('DFSDM1LPEN',          1),
                                            ('HRTIMLPEN',           1),
                                            ]),
            Reg32('C2_APB4LPENR',  0x1DC,  [('',                    1),
                                            ('SYSCFGLPEN',          1),
                                            ('',                    1),
                                            ('LPUART1LPEN',         1),
                                            ('',                    1),
                                            ('SPI6LPEN',            1),
                                            ('',                    1),
                                            ('I2C4LPEN',            1),
                                            ('',                    1),
                                            ('LPTIM2LPEN',          1),
                                            ('LPTIM3LPEN',          1),
                                            ('LPTIM4LPEN',          1),
                                            ('LPTIM5LPEN',          1),
                                            ('',                    1),
                                            ('COMP12LPEN',          1),
                                            ('VREFLPEN',            1),
                                            ('RTCAPBLPEN',          1),
                                            ('',                    4),
                                            ('SAI4LPEN',            1),
                                            ]),
            ]

    def __init__(self, target, ap, name, addr, **kwargs):
        super(RCC, self).__init__(target, ap, addr, name, RCC.REGS, **kwargs)

        self._f_hse = None

    @property
    def f_csi(self):
        return 4000000

    @property
    def f_hsi(self):
        while self._CR.HSIDIVF == 0:
            time.sleep(0.01)
        return (64000000 >> self._CR.HSIDIV)

    def set_f_hse(self, f):
        '''
        Called by external code that has knowledge of the target board through
        some other means.
        '''
        self._f_hse = f

    @property
    def f_hse(self):
        return self._f_hse

    @property
    def f_pllsrc(self):
        pllsrc = self._PLLCKSELR.PLLSRC
        if pllsrc == 0:
            return self.f_hsi
        if pllsrc == 1:
            return self.f_csi
        if pllsrc == 2:
            return self.f_hse
        return 0

    @property
    def f_pll1_p_clk(self):
        if self._CR.PLL1ON == 0:
            return 0
        if self._PLLCFGR.DIVP1EN == 0:
            return 0

        M = self._PLLCKSELR.DIVM1
        N = self._PLL1DIVR.DIVN1 + 1
        P = self._PLL1DIVR.DIVP1 + 1
        return self.f_pllsrc * N / (M * P)

    @property
    def f_sysclk(self):
        sws = self._CFGR.SWS
        if sws == 0:
            return self.f_hsi
        if sws == 1:
            return self.f_csi
        if sws == 2:
            return self.f_hse
        if sws == 3:
            return self.f_pll1_p_clk

    @property
    def f_sys_d1cpre_ck(self):
        d1cpre  = self._D1CFGR.D1CPRE
        divider = PRE_REV_MAP[d1cpre]
        return self.f_sysclk / divider

    @property
    def f_hclk(self):
        hpre    = self._D1CFGR.HPRE
        divider = PRE_REV_MAP[hpre]
        return self.f_sys_d1cpre_ck / divider

    @property
    def f_pclk1(self):
        d2ppre1 = self._D2CFGR.D2PPRE1
        divider = DPRE_REV_MAP[d2ppre1]
        return self.f_hclk / divider

    @property
    def f_pclk2(self):
        d2ppre2 = self._D2CFGR.D2PPRE2
        divider = DPRE_REV_MAP[d2ppre2]
        return self.f_hclk / divider

    @property
    def f_pclk3(self):
        d1ppre  = self._D1CFGR.D1PPRE
        divider = DPRE_REV_MAP[d1ppre]
        return self.f_hclk / divider

    @property
    def f_pclk4(self):
        d3ppre  = self._D3CFGR.D3PPRE
        divider = DPRE_REV_MAP[d3ppre]
        return self.f_hclk / divider

    @property
    def f_timx_ker_ck(self):
        d2ppre1 = self._D2CFGR.D2PPRE1
        timpre  = self._CFGR.TIMPRE
        divider = TIM_KER_DIVIDER[(timpre << 3) | d2ppre1]
        return self.f_hclk / divider

    @property
    def f_timy_ker_ck(self):
        d2ppre2 = self._D2CFGR.D2PPRE2
        timpre  = self._CFGR.TIMPRE
        divider = TIM_KER_DIVIDER[(timpre << 3) | d2ppre2]
        return self.f_hclk / divider

    def enable_device(self, name):
        offset, bit = ENABLE_BITS[name]
        if self._get_field(1, bit, offset) == 0:
            self._set_field(1, 1, bit, offset)

    def enable_hse(self):
        self._CR.HSEON = 1
        while self._CR.HSERDY == 0:
            time.sleep(0.01)

    def set_d1cpre(self, divider):
        bits                = PRE_MAP[divider]
        self._D1CFGR.D1CPRE = bits
        while self._D1CFGR.D1CPRE != bits:
            time.sleep(0.01)

    def set_hpre(self, divider):
        bits              = PRE_MAP[divider]
        self._D1CFGR.HPRE = bits
        while self._D1CFGR.HPRE != bits:
            time.sleep(0.01)

    def set_d1ppre(self, divider):
        bits                = DPRE_MAP[divider]
        self._D1CFGR.D1PPRE = bits
        while self._D1CFGR.D1PPRE != bits:
            time.sleep(0.01)

    def set_d2ppre1(self, divider):
        bits                 = DPRE_MAP[divider]
        self._D2CFGR.D2PPRE1 = bits
        while self._D2CFGR.D2PPRE1 != bits:
            time.sleep(0.01)

    def set_d2ppre2(self, divider):
        bits                 = DPRE_MAP[divider]
        self._D2CFGR.D2PPRE2 = bits
        while self._D2CFGR.D2PPRE2 != bits:
            time.sleep(0.01)

    def set_d3ppre(self, divider):
        bits                = DPRE_MAP[divider]
        self._D3CFGR.D3PPRE = bits
        while self._D3CFGR.D3PPRE != bits:
            time.sleep(0.01)

    def set_sysclock_source(self, sw):
        '''
        Selects the SYSCLOCK source as follows:

                SW | Source
                ---+-------
                0  | HSI (8, 16, 32 or 64 MHz)
                1  | CSI (4 MHz)
                2  | HSE
                3  | PLL
                ---+-------
        '''
        self._CFGR.SW = sw
        while self._CFGR.SWS != sw:
            time.sleep(0.01)

    def set_pll_source(self, pllsrc):
        '''
        Selects the PLL source as follows:

                PLLSRC | Source
                -------+-------
                0      | HSI
                1      | CSI
                2      | HSE
                3      | None (power saving)
                -------+-------
        '''
        self._PLLCKSELR.PLLSRC = pllsrc

    def configure_pll_p_clk(self, M, N, P):
        '''
        Given M, N, and P settings, configure PLL P CLK so that it could
        then be selected as the system clock.  Note that the VOS must already
        have been configured as appropriate for this setting by the client.

        This assumes we always use the wide VCO range.
        '''
        f_pllsrc = self.f_pllsrc
        f_pllref = f_pllsrc / M
        assert 1 <= M <= 63
        assert 4 <= N <= 512
        assert P and not (P % 2)
        assert 2000000 <= f_pllref <= 16000000

        self._PLLCKSELR.DIVM1 = M
        if f_pllref <= 4000000:
            self._PLLCFGR.PLL1RGE = 1
        elif f_pllref <= 8000000:
            self._PLLCFGR.PLL1RGE = 2
        else:
            self._PLLCFGR.PLL1RGE = 3
        self._PLLCFGR.PLL1VCOSEL = 0
        self._PLLCFGR.PLL1FRACEN = 0
        self._PLL1DIVR.DIVN1     = N - 1
        self._PLL1DIVR.DIVP1     = P - 1
        self._PLL1FRACR          = 0
        self._PLLCFGR.DIVP1EN    = 1
        self._CR.PLL1ON          = 1
        while self._CR.PLL1RDY == 0:
            time.sleep(0.01)

    @staticmethod
    def get_pll_mnpv(f_target_mhz, f_hsclk):
        '''
        Given an input of frequency f_hsclk to the M prescaler, compute the
        M, N, P values to get an output frequency of f_target_mhz.  The target
        value should be an integral value in MHz (not Hz!) while f_hsclk should
        be in Hz.  We compute M, N, P such that:

            f_target_mhz = f_hsclk_mhz * N / (M * P)

        We also compute the VOS mode.

        We have the following general constraints:

            1 <= M <= 63

            f_pllin_mhz = f_hsclk_mhz / M
            2 <= f_pllin_mhz <= 16      (assuming wide VCO range)

            f_vcoout_mhz = f_pllin_mhz * N
            4 <= N <= 512

            f_target_mhz = f_vcoout_mhz / P
            P in {2, 4, 6, 8, 10, 12, ..., 128}

        We have the following constraints in wide VCO frequency range (used
        when the output from the M prescaler is in the range 2 - 16 MHz):

            VOS0:   1.5 <= f_target_mhz <= 480
            VOS1:   1.5 <= f_target_mhz <= 400
            VOS2:   1.5 <= f_target_mhz <= 300
            VOS3:   1.5 <= f_target_mhz <= 200
            192 <= f_vcoout_mhz <= 960

        We have the following constraints in medium VCO frequency range (used
        when the output from the M prescaler is in the range 1 - 2 MHz):

            VOS1:   1.17 <= f_target_mhz <= 210
            VOS2:   1.17 <= f_target_mhz <= 210
            VOS3:   1.17 <= f_target_mhz <= 200
            150 <= f_vcoout_mhz <= 420

        We'll assume for now we use the wide range.
        '''
        assert 2 <= f_target_mhz <= 480

        # First find the optimal voltage setting.
        if f_target_mhz <= 200:
            vos = 3
        elif f_target_mhz <= 300:
            vos = 2
        elif f_target_mhz <= 400:
            vos = 1
        elif f_target_mhz <= 480:
            vos = 0

        # Find the range of allowed M settings, restricting ourselves so that
        # f_pllin_mhz is between 2 - 16 MHz.
        M_min = max(1, math.ceil(f_hsclk / 16000000))
        M_max = min(63, math.floor(f_hsclk / 2000000))

        # Brute force the rest of it.  We attempt to minimize the
        # multiplicative factor N under the assumption that more multiplication
        # means more jitter.  Then we try to maximize the M prescaler in order
        # to lower the VCO frequency and thus reduce power consumption.
        f_hsclk_mhz = math.floor(f_hsclk / 1000000)
        for N in range(4, 512):
            for M in range(M_max, M_min - 1, -1):
                f_vcoout_mhz = f_hsclk_mhz * N // M
                if 192 <= f_vcoout_mhz <= 960:
                    P = f_vcoout_mhz // f_target_mhz
                    if not P or P % 2:
                        continue
                    if (f_hsclk_mhz * N) % (M * P):
                        continue
                    if f_hsclk_mhz * N // (M * P) == f_target_mhz:
                        return M, N, P, vos

        raise Exception('No valid M, N, P combination!')
